#!/usr/bin/tclsh

# This tool generates Tcl code from an 'odl' schema.

# Data structures used by the tool:
# (x = module, y = class)
# - list odl_modulenames= { nam1 nam2 ... }
# - odl_modules(x,paramlist)= { {nam val} {nam val} ... }
# - odl_modules(x,classlist)= { nam1 nam2 ... }
# - odl_classes(x,y,parentlist)= { nam1 nam2 ... }
# - odl_classes(x,y,attrlist)= { {nam val} {nam val} ... }
# - odl_classes(x,y,methodlist)= { {cls mod nam params} ... }
# - odl_classes(x,y,containslist)= { {cls inst} {cls inst} ... }

# Debugging output; switch it on by removing the comment.
proc deb {msg} {
#   puts $msg
}

# Error output
proc odl_err { msg } {
   puts "ERR: $msg"
}

# List of all modules in the schema (usually just 1)
set odl_modulenames [list]

#######
# SYN procs
# Execute the 'odl' schema and record its information in
# internal data structures.

set odl_cur_modulename ""
set odl_cur_classname ""

proc odl_module { name body } {
   global odl_modulenames

   if { [lsearch $odl_modulenames $name] != -1 } {
      odl_err "Double definition of module $name (ignored)"
      return
   }

   lappend odl_modulenames $name

   global odl_modules
   set odl_modules($name,paramlist) [list]
   set odl_modules($name,classlist) [list]

   # Remember what module we're currently in
   global odl_cur_modulename
   set odl_cur_modulename $name

   # Evaluate module body which in turn contains class definitions
   uplevel 1 $body

   # Now we're not in a module anymore
   set odl_cur_modulename ""
}

proc param { name value } {
   global odl_modules
   global odl_cur_modulename

   lappend odl_modules($odl_cur_modulename,paramlist) [list $name $value]
}

proc class { name parentlist body } {
   global odl_modules
   global odl_cur_modulename

   set classlist $odl_modules($odl_cur_modulename,classlist)
   if { [lsearch $classlist $name] != -1 } {
      odl_err "Double definition of class $name (ignored)"
      return
   }

   lappend classlist $name
   set odl_modules($odl_cur_modulename,classlist) $classlist

   global odl_classes
   set odl_classes($odl_cur_modulename,$name,parentlist) $parentlist
   set odl_classes($odl_cur_modulename,$name,attrlist) [list]
   set odl_classes($odl_cur_modulename,$name,methodlist) [list]
   set odl_classes($odl_cur_modulename,$name,containslist) [list]

   global odl_cur_classname
   set odl_cur_classname $name
   uplevel 1 $body
   set odl_cur_classname ""
}

proc contains { classname instancename } {
   global odl_classes
   global odl_cur_modulename
   global odl_cur_classname

   lappend odl_classes($odl_cur_modulename,$odl_cur_classname,containslist) \
      [list $classname $instancename]
}

proc attr { name defaultvalue } {
   global odl_classes
   global odl_cur_modulename
   global odl_cur_classname

   lappend odl_classes($odl_cur_modulename,$odl_cur_classname,attrlist) \
      [list $name $defaultvalue]
}

proc method { modifiers name paramlist } {
   global odl_classes
   global odl_cur_modulename
   global odl_cur_classname

   lappend odl_classes($odl_cur_modulename,$odl_cur_classname,methodlist) \
      [list $odl_cur_classname $modifiers $name $paramlist]
}

#######
# Some helper procs
#

proc new_class_name { cls new_name } {
   if { $new_name == "" } {
      return $cls
   }
   if { $new_name == "-" } {
      return $cls
   }
   return "${cls}_${new_name}"
}

#######
# Main output generation proc.
# Generates the code and a skeleton file.

proc output_procs { procnam skelnam } {

   set fodp [open $procnam "w"]
   set fods [open $skelnam "w"]

   global argv

   puts $fodp "### This file was generated by tcl_odl from '[lindex $argv 0]'"
   puts $fodp "###"
   puts $fodp "### It contains procs to create and delete objects,"
   puts $fodp "### and to create local variables for their members."
   puts $fodp "###"

   puts $fods "### This file was generated by tcl_odl from '[lindex $argv 0]'"
   puts $fods "###"
   puts $fods "### It contains skeletons for the methods of all classes."
   puts $fods "### You still need to implement those methods manually."
   puts $fods "###"

   global odl_modulenames
   global odl_modules
   global odl_classes

   foreach mod $odl_modulenames {

      # Output module header
      puts $fodp ""
      puts $fodp "### Module $mod"
      puts $fodp ""

      # Set module params, default values
      set all [list]
      set new_name ""
      set delete_name "del_$mod"
      set member_name "mem_$mod"
      set init_name "-"
      set exit_name "-"
      set get_name "-"
      set set_name "-"
      set debug 0

      # Set module params, actual values
      foreach param $odl_modules($mod,paramlist) {
         set [lindex $param 0] [lindex $param 1]
      }

      # Some debugging output
      deb "module $mod:"
      deb "  all= $all"
      deb "  new_name= $new_name"
      deb "  delete_name= $delete_name"
      deb "  member_name= $member_name"
      deb "  init_name= $init_name"
      deb "  exit_name= $exit_name"
      deb "  get_name= $get_name"
      deb "  set_name= $set_name"
      deb "  debug= $debug"

      foreach cls $odl_modules($mod,classlist) {

         set parentlist $odl_classes($mod,$cls,parentlist)
         set attrlist $odl_classes($mod,$cls,attrlist)
         set methodlist $odl_classes($mod,$cls,methodlist)
         set containslist $odl_classes($mod,$cls,containslist)

         # Add all attrs and methods of all parent classes.
         # We only do the immediate parents; recursive parents
         # are handled below.
         foreach parent $parentlist {
            set tmp $odl_classes($mod,$parent,attrlist)
            foreach elt $tmp {
               lappend attrlist $elt
            }

            set tmp $odl_classes($mod,$parent,methodlist)
            foreach elt $tmp {
               # If the method has the same name as an already existing one,
               # it overrides the existing one.  Otherwise, it's just a
               # new method.
               set found 0
               set m_name [lindex $elt 2]
               foreach existing $methodlist {
                  if { [lindex $existing 2] == $m_name } {
                     set found 1
                     break
                  }
               }

               if { !$found } {
                  lappend methodlist $elt
               }
            }

            set tmp $odl_classes($mod,$parent,containslist)
            foreach elt $tmp {
               lappend containslist $elt
            }
         }

         # Recursive inheritance -> store information for later
         # subclasses.  Original schema information is lost.  So what.
         set odl_classes($mod,$cls,attrlist) $attrlist
         set odl_classes($mod,$cls,methodlist) $methodlist
         set odl_classes($mod,$cls,containslist) $containslist

         # Add the 'all' methods of the module
         # (all classes automatically get these methods).
         foreach method $all {
            lappend methodlist [list $cls "" $method "args"]
         }

         # Some more debugging output
         deb "  -- class $cls: $parentlist"
         deb "     attrlist= $attrlist"
         deb "     methodlist= $methodlist"
         deb "     containslist= $containslist"

         ###
         # Generate the method dispatchers (one per class)

         puts $fodp "\nproc objdisp_${cls} {handle cmd args} \{"
         puts $fodp "   upvar #0 \${handle}_array local_array"
         puts $fodp ""
         puts $fodp "   if \{ \$cmd == \"configure\" \} \{"
         puts $fodp "      foreach \{ opt val \} \$args \{"
         puts $fodp "         if \{ !\[regexp \{^-(.+)\} \$opt dummy small_opt\] \} \{"
#         puts $fodp "            puts \"Wrong option name \$opt (ignored)\""
         puts $fodp "         \} else \{"
         if { $set_name != "-" } {
         puts $fodp "            set old_value \$local_array(\$small_opt)"
         }
         puts $fodp "            set local_array(\$small_opt) \$val"
         if { $set_name != "-" } {
         puts $fodp "            ${cls}_$set_name \$handle \$small_opt \$old_value \$val"
         }
         puts $fodp "         \}"
         puts $fodp "      \}"
         puts $fodp "      return 1"
         puts $fodp "   \}"
         puts $fodp ""
         puts $fodp "   if \{ \$cmd == \"cget\" \} \{"
         puts $fodp "      set opt \[lindex \$args 0\]"
         puts $fodp "      if \{ !\[regexp \{^-(.+)\} \$opt dummy small_opt\] \} \{"
         puts $fodp "         puts \"Wrong or missing option name \$opt\""
         puts $fodp "         return \"\""
         puts $fodp "      \}"
         if { $get_name != "-" } {
         puts $fodp "      ${cls}_$get_name \$handle \$small_opt"
         }
         puts $fodp "      return \$local_array(\$small_opt)"
         puts $fodp "   \}"

         foreach method $methodlist {
            set c [lindex $method 0]
            set m [lindex $method 2]
            puts $fodp ""
            puts $fodp "   if \{ \$cmd == \"$m\" \} \{"
            puts $fodp "      if \{ \[ llength \$args \] == 0 \} \{"
            puts $fodp "         return \[ eval ${c}_${m} \$handle \]"
            puts $fodp "      \} else \{"
            puts $fodp "         return \[ eval ${c}_${m} \$handle \$args \]"
            puts $fodp "      \}"
            puts $fodp "   \}"
         }

         puts $fodp ""
         puts $fodp "   puts \"Unknown command \$cmd\""
         puts $fodp "\}"

         ###
         # Generate the 'new' procs (one per class)

         puts $fodp "\nproc [new_class_name $cls $new_name] {handle args} \{"

         if { $debug } {
            puts $fodp "   puts \"CTOR $cls \$handle\""
         }

         puts $fodp "   # Remember the object's type:"
         puts $fodp "   global odl_types_array"
         puts $fodp "   set odl_types_array(\$handle) ${cls}"
         puts $fodp ""

         puts $fodp "   # Install the array of object parameters:"
         puts $fodp "   global \$\{handle\}_array"
         foreach attr $attrlist {
            puts $fodp "   set \$\{handle\}_array([lindex $attr 0]) \"[lindex $attr 1]\""
         }
         puts $fodp ""
         puts $fodp ""

         puts $fodp "   # Install a proc that will serve as the \"object command\":"
         puts -nonewline $fodp "   proc \$handle {cmd args} \""
         puts -nonewline $fodp "return \\\[eval objdisp_${cls} \$handle \\\$cmd \\\$args\\\]"
         puts $fodp "\""

         if { [llength $containslist] > 0 } {
            puts $fodp ""
            puts $fodp "   # Construct the contained objects as well:"
            foreach contained $containslist {
               set tmp [new_class_name [lindex $contained 0] $new_name]
               puts -nonewline $fodp "   $tmp"
               puts $fodp " \$\{handle\}.[lindex $contained 1]"
            }
         }

         puts $fodp ""
         puts $fodp "   # Now configure using the other args:"
         puts $fodp "   eval \$handle configure \$args"

         if { $init_name != "-" } {
            puts $fodp ""
            puts $fodp "   # Call the object's initialization proc"
            puts $fodp "   ${cls}_${init_name} \$handle"
         }

         puts $fodp "\}\n\n"

         ###
         # Generate the skeleton implementations for methods

         puts $fods ""
         puts $fods "#"
         puts $fods "# $cls"
         puts $fods "#"
         puts $fods ""

         foreach method $methodlist {
            set c [lindex $method 0]
            if { $c == $cls } {
               puts -nonewline $fods "proc ${cls}_[lindex $method 2] \{ handle "
               foreach param [lindex $method 3] {
                  if { $param == "args" } {
                     puts -nonewline $fods "args "
                  } else {
                     if { [llength $param] > 1 } {
                        puts -nonewline $fods "\{[lindex $param 0]"
                        puts -nonewline $fods " [lindex $param 1]\} "
                     } else {
                        puts -nonewline $fods "[lindex $param 0] "
                     }
                  }
               }
               puts $fods "\} \{\n\}\n\n"
            }
         }

         foreach method [list $init_name $exit_name] {
            if { $method != "-" } {
               puts $fods "proc ${cls}_${method} \{ handle \} \{\n\}\n\n"
            }
         }
         if { $get_name != "-" } {
            puts $fods "proc ${cls}_${get_name} \{ handle opt \} \{\n\}\n\n"
         }
         if { $set_name != "-" } {
            puts $fods "proc ${cls}_${set_name} \{ handle opt oldval newval \} \{\n\}\n\n"
         }

         ###
         # Generate the 'delete' proc (one per class)

         if { $delete_name != "-" } {
            puts $fodp "\nproc ${cls}_${delete_name} {handle} \{"

            if { $debug } {
               puts $fodp "   puts \"DTOR $cls \$handle\""
            }

            if { $exit_name != "-" } {
               puts $fodp ""
               puts $fodp "   # Call the object's exit proc"
               puts $fodp "   ${cls}_${exit_name} \$handle"
            }

            foreach contained $containslist {
               puts -nonewline $fodp "   [lindex $contained 0]_${delete_name}"
               puts $fodp " \$\{handle\}.[lindex $contained 1]"
            }
            puts $fodp "   catch { rename \$handle {} }"
            puts $fodp "   global \$\{handle\}_array"
            puts $fodp "   unset \$\{handle\}_array"
            puts $fodp "\}\n\n"
         }

      }   ; # foreach cls

      ###
      # Generate the 'delete' proc (one per module)
      if { $delete_name != "-" } {
         puts $fodp "\nproc $delete_name {args} \{"
         puts $fodp "   global odl_types_array"
         puts $fodp "   foreach handle \$args \{"
         puts $fodp "      \$odl_types_array(\$handle)_${delete_name} \$handle"
         puts $fodp "   \}"
         puts $fodp "\}\n\n"
      }

      ###
      # Generate the 'member' proc (one per module)
      if { $member_name != "-" } {
         puts $fodp "\nproc $member_name {handle args} \{"
         puts $fodp "   upvar #0 \$\{handle\}_array local_array"
         puts $fodp "   foreach nam \$args \{"
         puts $fodp "      upvar \$nam tmp"
         puts $fodp "      set tmp \$local_array(\$nam)"
         puts $fodp "   \}"
         puts $fodp "\}\n\n"
      }

   }

   close $fodp
   close $fods
}

#######
# "main"

if { [llength $argv] != 3 } {
   puts "Usage: tcl_odl.tcl <schema file> <procs target> <skeletons target>"
} else {
   source [lindex $argv 0]
   output_procs [lindex $argv 1] [lindex $argv 2]
}
