# displayer.tcl

# Create displayer canvas with two scrollbars
proc scrollcanvas {parent_name} {

  set f1 [frame $parent_name.top -relief solid  -borderwidth 2]
  pack $f1 -side top -expand true -fill both

  set result $f1.cnv
  set scrx \
    [scrollbar $f1.xsbar -orient horizontal -command "$result xview"]
  set scry \
    [scrollbar $f1.ysbar -orient vertical -command "$result yview"]
  canvas $result \
    -background bisque \
    -xscrollcommand "$scrx set" \
    -yscrollcommand "$scry set"

  grid $result -row 0 -column 0 -sticky nsew
  grid $scry -row 0 -column 1 -sticky ns
  grid $scrx -row 1 -column 0 -sticky ew

  grid columnconfigure $f1 0 -weight 1
  grid rowconfigure $f1 0 -weight 1

  return $result
}

# Currently active displayer (used in all the d_... procs)
set displayer ""

# Displayer font for all text
set d_defaultFont "-adobe-courier-bold-r-normal--12-120-75-75-m-70-iso8859-1"

set a4_pagewidth 600
set a4_pageheight 850

proc d_Clear {} {
  global displayer
  $displayer delete all
}


proc d_LoadFile {fileName} {
  source $fileName
}

################################################################################
# Basic displayer procs
################################################################################

###
###  Rectangles and text
###

proc d_rect {x y w h o c tgs} {
  global displayer
  if {$c == "none"} {
    $displayer create line $x $y [expr $x + $w - 1] $y \
      [expr $x + $w - 1] [expr $y + $h - 1] $x [expr $y + $h - 1] \
      $x $y \
      -width 3 \
      -fill $o \
      -tags $tgs
  } else {
    $displayer create rectangle $x $y [expr $x + $w - 1] [expr $y + $h - 1] \
      -outline $o \
      -fill $c \
      -tags $tgs
  }
}

proc d_txt {x y t o a tgs} {
  global d_defaultFont
  global displayer

  $displayer create text $x $y \
    -anchor $a \
    -fill $o \
    -tags $tgs \
    -text $t \
    -font $d_defaultFont
}

proc d_namedRect {x y w h clr txt txtclr tgs} {
  d_rect $x $y $w $h $txtclr $clr $tgs
  d_txt [expr $x + [expr $w / 2]] [expr $y + [expr $h / 2]] $txt $txtclr c $tgs
}

proc d_namedLeftRect {x y w h margin clr txt txtclr tgs} {
  d_rect $x $y $w $h $txtclr $clr $tgs
  d_txt [expr $x + $margin] [expr $y + $margin] $txt $txtclr nw $tgs
}

proc d_doubleRect {x y w h o c tgs} {
  global displayer
  if {$c == "none"} {
    $displayer create rectangle \
      [expr $x - 1] [expr $y - 1] \
      [expr $x + $w] [expr $y + $h] \
      -outline $o \
      -tags $tgs
    $displayer create rectangle \
      [expr $x + 1] [expr $y + 1] \
      [expr $x + $w - 2] [expr $y + $h - 2] \
      -outline $o \
      -tags $tgs
  } else {
    $displayer create rectangle \
      [expr $x - 1] [expr $y - 1] \
      [expr $x + $w] [expr $y + $h] \
      -outline $o \
      -fill $c \
      -tags $tgs
    $displayer create rectangle \
      [expr $x + 1] [expr $y + 1] \
      [expr $x + $w - 2] [expr $y + $h - 2] \
      -outline $o \
      -fill $c \
      -tags $tgs
  }
}

proc d_diamond {x y w h o c tgs} {
  global displayer
  if {$c == "none"} {
    $displayer create line \
      [expr $x + $w/2] $y \
      [expr $x + $w] [expr $y + $h/2] \
      [expr $x + $w/2] [expr $y + $h] \
      $x [expr $y + $h/2] \
      [expr $x + $w/2] $y \
      -fill $o \
      -width 1 \
      -tags $tgs
  } else {
    $displayer create polygon \
      [expr $x + $w/2] $y \
      [expr $x + $w] [expr $y + $h/2] \
      [expr $x + $w/2] [expr $y + $h] \
      $x [expr $y + $h/2] \
      [expr $x + $w/2] $y \
      -fill $c -outline $o \
      -width 1 \
      -tags $tgs
  }
}

###
###  Vectors and matrices
###

proc d_row {x y w h n o c tgs} {
  set row_curX $x
  set row_W [expr $w - 1]
  for {set i 0} {$i < $n} {incr i} {
    d_rect $row_curX $y $w $h $o $c $tgs
    incr row_curX $row_W
  }
}

proc d_col {x y w h n o c tgs} {
  set col_curY $y
  set col_H [expr $h - 1]
  for {set i 0} {$i < $n} {incr i} {
    d_rect $x $col_curY $w $h $o $c $tgs
    incr col_curY $col_H
  }
}

proc d_clrrow {x y w h o c tgs} {
  set row_curX $x
  set row_W [expr $w - 1]
  set n [llength $o]
  for {set i 0} {$i < $n} {incr i} {
    d_rect $row_curX $y $w $h [lindex $o $i] [lindex $c $i] $tgs
    incr row_curX $row_W
  }
}

proc d_clrcol {x y w h o c tgs} {
  set col_curY $y
  set col_H [expr $h - 1]
  set n [llength $o]
  for {set i 0} {$i < $n} {incr i} {
    d_rect $x $col_curY $w $h [lindex $o $i] [lindex $c $i] $tgs
    incr col_curY $col_H
  }
}

proc d_txtrow {x y w h tList o c txtclr tgs} {
  set n [llength $tList]
  d_row $x $y $w $h $n $o $c $tgs

  set tr_curX [expr $x + [expr $w / 2]]
  set tr_curY [expr $y + [expr $h / 2]]
  foreach txt $tList {
    if { $txt != "" } {
      d_txt $tr_curX $tr_curY $txt $txtclr c $tgs
    }
    incr tr_curX $w
  }
}

proc d_txtcol {x y w h tList o c txtclr tgs} {
  set n [llength $tList]
  d_col $x $y $w $h $n $o $c $tgs

  set tr_curX [expr $x + [expr $w / 2]]
  set tr_curY [expr $y + [expr $h / 2]]
  foreach txt $tList {
    if { $txt != "" } {
      d_txt $tr_curX $tr_curY $txt $txtclr c $tgs
    }
    incr tr_curY $h
  }
}

proc d_clrtxtrow {x y w h tList o c txtclr tgs} {
  set n [llength $tList]
  d_clrrow $x $y $w $h $o $c $tgs

  set tr_curX [expr $x + [expr $w / 2]]
  set tr_curY [expr $y + [expr $h / 2]]
  for {set i 0} {$i < $n} {incr i} {
    set txt [lindex $tList $i]
    if { $txt != "" } {
      d_txt $tr_curX $tr_curY $txt [lindex $txtclr $i] c $tgs
    }
    incr tr_curX $w
  }
}

proc d_clrtxtcol {x y w h tList o c txtclr tgs} {
  set n [llength $tList]
  d_clrcol $x $y $w $h $o $c $tgs

  set tr_curX [expr $x + [expr $w / 2]]
  set tr_curY [expr $y + [expr $h / 2]]
  for {set i 0} {$i < $n} {incr i} {
    set txt [lindex $tList $i]
    if { $txt != "" } {
      d_txt $tr_curX $tr_curY $txt [lindex $txtclr $i] c $tgs
    }
    incr tr_curY $h
  }
}

proc d_grid {x y w h m n o c tgs} {
  set grid_W [expr $w - 1]
  set grid_H [expr $h - 1]
  set grid_curX $x
  for {set i 0} {$i < $m} {incr i} {
    set grid_curY $y
    for {set j 0} {$j < $n} {incr j} {
      d_rect $grid_curX $grid_curY $w $h $o $c $tgs
      incr grid_curY $grid_H
    }
    incr grid_curX $grid_W
  }
}


###
### Lines and arrows
###

proc d_line {x1 y1 x2 y2 w o tgs} {
  global displayer

  $displayer create line $x1 $y1 $x2 $y2 \
    -width $w \
    -fill $o \
    -tags $tgs
}

proc d_arrow {x1 y1 x2 y2 w o tgs} {
  global displayer

  $displayer create line $x1 $y1 $x2 $y2 \
    -width $w \
    -fill $o \
    -tags $tgs \
    -arrow last
}

proc d_arrow2 {x1 y1 x2 y2 w o tgs} {
  global displayer

  $displayer create line $x1 $y1 $x2 $y2 \
    -width $w \
    -fill $o \
    -tags $tgs \
    -arrow both
}

###
### Circles and ovals
###

# Automatically centered oval
proc d_oval {x y w h o c tgs} {
  global displayer

  $displayer create oval \
    [expr $x - ($w/2)] [expr $y - ($h/2)] \
    [expr $x + ($w/2)] [expr $y + ($h/2)] \
    -fill $c -outline $o \
    -tags $tgs
}

proc d_namedOval {x y w h clr txt txtclr tgs} {
  d_oval $x $y $w $h $txtclr $clr $tgs
  d_txt $x $y $txt $txtclr c $tgs
}

# oval specified by its bounding box
proc d_BoxOval {x y w h o c tgs} {
  global displayer

  $displayer create oval \
    $x $y [expr $x + $w] [expr $y + $h] \
    -fill $c -outline $o \
    -tags $tgs
}

proc d_namedBoxOval {x y w h clr txt txtclr tgs} {
  d_BoxOval $x $y $w $h $txtclr $clr $tgs
  d_txt [expr $x + ($w/2)] [expr $y + ($h/2)] $txt $txtclr c $tgs
}


###
### More complex figures
###

proc d_disk {x y w h o c tgs} {
  d_BoxOval $x [expr $y + 3*($h/4)] $w [expr $h/4] $o $c $tgs
  d_rect $x [expr $y + ($h/8)] $w [expr $h - ($h/4)] $c $c $tgs
  d_BoxOval $x $y $w [expr $h/4] $o $c $tgs

  d_line $x [expr $y + ($h/8)] $x [expr $y + $h - ($h/8)] 1 $o $tgs
  d_line [expr $x + $w] [expr $y + ($h/8)] \
         [expr $x + $w] [expr $y + $h - ($h/8)] \
         1 $o $tgs
}

proc d_namedDisk {x y w h o c txt txtclr tgs} {
  d_disk $x $y $w $h $o $c $tgs
  d_txt [expr $x + ($w/2)] [expr $y + 7*($h/8)] $txt $txtclr c $tgs
}

proc d_file {x y w h o c tgs} {
  global displayer

  $displayer create polygon \
    [expr $x + ($w/8)] $y \
    [expr $x + $w] $y \
    [expr $x + $w] [expr $y + $h] \
    $x [expr $y + $h] \
    $x [expr $y + ($h/8)] \
    [expr $x + ($w/8)] $y \
    -fill $c -outline $o \
    -width 1 \
    -tags $tgs
  d_line $x [expr $y + ($h/8)] [expr $x + ($w/8)] [expr $y + ($h/8)] \
    1 $o $tgs
  d_line [expr $x + ($w/8)] [expr $y + ($h/8)] [expr $x + ($w/8)] $y \
    1 $o $tgs
}

proc d_namedFile {x y w h o c txt txtclr tgs} {
  d_file $x $y $w $h $o $c $tgs
  d_txt [expr $x + ($w/2)] [expr $y + ($h/2)] $txt $txtclr c $tgs
}

proc d_nodes {x y w h o c tgs} {

  d_line [expr $x + ($w/2)] [expr $y + ($h/4)] \
         [expr $x + ($w/4)] [expr $y + 3*($h/4)] \
         1 $o $tgs
  d_line [expr $x + ($w/2)] [expr $y + ($h/4)] \
         [expr $x + 3*($w/4)] [expr $y + 3*($h/4)] \
         1 $o $tgs

  d_oval [expr $x + ($w/2)] [expr $y + ($h/4)] [expr $w/4] [expr $h/4] \
    $o $c $tgs
  d_oval [expr $x + ($w/4)] [expr $y + 3*($h/4)] [expr $w/4] [expr $h/4] \
    $o $c $tgs
  d_oval [expr $x + 3*($w/4)] [expr $y + 3*($h/4)] [expr $w/4] [expr $h/4] \
    $o $c $tgs
}

# Create displayer canvas with two scrollbars

wm geometry . 980x600+15+5
wm title    . "Displayer"

set displayer [scrollcanvas .]

#set d_defaultFont "-*-Screen-Medium-R-*-*-*-160-*-*-*-*-*-*"
#set d_defaultFont "screen.r.9"
